#####################################################################
# Auxiliary functions for the kohonen map and other implementations #
# Ricardo Cerri - 02/07/2019                                        # 
#####################################################################


# n.neurons.grid was obtained with 40% of training data
# Let's get the corresponding number for the whole dataset, using a simple rule
adjustNeuronsGrid <- function(n.neurons.grid,average.num.instances,offline.classes){
  
  new.n.neurons.grid <- vector()
  
  for(i in 1:length(n.neurons.grid)){
    
    # Total number of instances from class i
    total.i <- length(which(offline.classes[,i]==1))
    
    # Compute adjusted number (simple rule)
    n.grid <- (total.i * n.neurons.grid[i]) / average.num.instances[i] 
    n.grid <- floor(n.grid)
    
    new.n.neurons.grid <- c(new.n.neurons.grid,n.grid)
  }
  
  return(new.n.neurons.grid)
}


# Given a micro-cluster, calculates the largest distance between its centroid (weight-vector) and 
# the centroids of all micro-clusters from a given class
# -------------------------------------------------------------------------------------
getLargestDistancesMCclass <- function(som.micro.clusters,id.class,id.mc){
  
  dmax <- 0
  
  if(length(som.micro.clusters[[id.class]][[2]]) > 1){
    
    centroid1 <- som.micro.clusters[[id.class]][[2]][[id.mc]][[1]]
    dmaxi <- vector()
    
    for(i in 1:length(som.micro.clusters[[id.class]][[2]])){
      centroid2 <- som.micro.clusters[[id.class]][[2]][[i]][[1]]
      distance <- euc.dist(centroid1,centroid2)
      dmaxi <- c(dmaxi,distance)
    }
    
    #dmax <- max(dmaxi)
    dmax <- mean(dmaxi)
  }
  else{
    dmax <- som.micro.clusters[[id.class]][[2]][[id.mc]][[9]]
  }
  
  return(dmax)
}
# -------------------------------------------------------------------------------------

# Given the index of a class, calculates the higher distance between a centroid and the
# farthest instance from this centroid among all micro-cluster representing the class index (radius 1)
# and also the higher distance between a centroid and the farthest instance from the isolated one (radius 2)
# -------------------------------------------------------------------------------------
getLargestDistancesNeuron <- function(som.micro.clusters,id.class){
  
  maxRadius1Vet <- vector()
  maxRadius2Vet <- vector()
  
  for(id.mc in 1:length(som.micro.clusters[[id.class]][[2]])){
    maxRadius1Vet <- c(maxRadius1Vet,som.micro.clusters[[id.class]][[2]][[id.mc]][[9]])
    maxRadius2Vet <- c(maxRadius2Vet,som.micro.clusters[[id.class]][[2]][[id.mc]][[12]])
  }
  
  maxRadius1 <- max(maxRadius1Vet)
  maxRadius2 <- min(maxRadius2Vet)
  
  return(c(maxRadius1,maxRadius2))
}
# -------------------------------------------------------------------------------------

# Sort winning neurons and return structure with corresponding class ids and distances
# -------------------------------------------------------------------------------------------------
sort.winning.neurons <- function(distances,vet.distances,vet.indexes,n.k,num.patterns,winner,inst.l){
  
  prob.sorted <- list()
  prob.sorted$x <- vector()
  prob.sorted$ix <- vector()
  
  for(i in 1:num.patterns){
    vet.distances.ordered <- sort(vet.distances,index.return=TRUE)
    table.count.indexes <- table(vet.indexes[vet.distances.ordered$ix][1:n.k])
    
    pos <- as.numeric(names(table.count.indexes[names(table.count.indexes) %in% prob.sorted$ix]))
    if(length(pos) > 0){
      table.count.indexes.names <- as.numeric(names(table.count.indexes))
      table.count.indexes <- table.count.indexes[! table.count.indexes.names %in% pos]
    }
    
    if(length(table.count.indexes) > 0){
      id.max <- as.numeric(names(which(table.count.indexes == max(table.count.indexes))))
      
      if(length(id.max) > 1){ # Is there a tie?
        means.id.max <- vector()
        for(id in id.max){
          means.id.max <- c(means.id.max,mean(winner[[id]]$nn.dist[inst.l,1:n.k]))
        }
        id.max.mean <- sort(means.id.max,index.return=TRUE)
        id.max <- id.max[id.max.mean$ix[1]]
      }
      
      prob.sorted$ix <- c(prob.sorted$ix,id.max)
      prob.sorted$x <- c(prob.sorted$x,distances[id.max])
      vet.distances[vet.indexes==id.max] <- 999999
    }
  }
  
  
  if(length(prob.sorted$ix) < num.patterns){ # There is a class that never won
    all.classes <- c(1:num.patterns)
    last.class <- all.classes[! all.classes %in% prob.sorted$ix]
    prob.sorted$ix <- c(prob.sorted$ix,last.class)
    prob.sorted$x <- c(prob.sorted$x,distances[last.class])
  }
  
  if(length(prob.sorted$ix) < num.patterns){browser()}
  
  
  #---------------------------------------------------------------------------------------------------
#   vet.distances.ordered <- sort(vet.distances,index.return=TRUE)
#   
#   for(i in seq(from=1, to=length(vet.indexes), by=n.k)){
#     
#     if((i+n.k) > length(vet.indexes)){
#       table.count.indexes <- table(vet.indexes[vet.distances.ordered$ix][i:length(vet.indexes)])
#     }
#     else{
#       table.count.indexes <- table(vet.indexes[vet.distances.ordered$ix][i:(i+(n.k-1))])
#     }
#     pos <- as.numeric(names(table.count.indexes[names(table.count.indexes) %in% prob.sorted$ix]))
#     
#     if(length(pos) > 0){
#       table.count.indexes.names <- as.numeric(names(table.count.indexes))
#       table.count.indexes <- table.count.indexes[! table.count.indexes.names %in% pos]
#     }
#     
#     browser()
#     
#     if(length(table.count.indexes) > 0){
#       id.max <- as.numeric(names(which(table.count.indexes == max(table.count.indexes))))
#       
#       if(length(id.max) > 1){ # Is there a tie?
#         means.id.max <- vector()
#         for(id in id.max){
#           means.id.max <- c(means.id.max,mean(winner[[id]]$nn.dist[inst.l,1:n.k]))
#         }
#         id.max.mean <- sort(means.id.max,index.return=TRUE)
#         id.max <- id.max[id.max.mean$ix[1]]
#       }
#       
#       prob.sorted$ix <- c(prob.sorted$ix,id.max)
#       prob.sorted$x <- c(prob.sorted$x,distances[id.max])
#     }
#   }
#   
#   if(length(prob.sorted$ix) < num.patterns){ # There is a class that never won
#     all.classes <- c(1:num.patterns)
#     last.class <- all.classes[! all.classes %in% prob.sorted$ix]
#     prob.sorted$ix <- c(prob.sorted$ix,last.class)
#     prob.sorted$x <- c(prob.sorted$x,distances[last.class])
#   }
#   
#   if(length(prob.sorted$ix) < num.patterns){browser()}
  #---------------------------------------------------------------------------------------------------
  
  
  return(prob.sorted)
}
# -------------------------------------------------------------------------------------------------

# Compute initial class totals and class probabilities
# -------------------------------------------------------------------------------------------------
computeInitialClassProbabilitiesTotals <- function(offline.classes){
  
  result <- list()
  num.classes <- ncol(offline.classes) # number of initial classes. 
  
  # Fill class probabilities matrix
  # Also fill the total counts of examples per class and intersections.
  # This information will be used to further update class probabilities
  class.probabilities <- matrix(nrow=num.classes, ncol=num.classes, data=0)
  class.totals <- matrix(nrow=num.classes, ncol=num.classes, data=0)
  for(i in 1:num.classes){
    for(j in 1:num.classes){
      class.prob.totals <- get.probabilities(offline.classes,i,j)  # p(i) or p(i|j) and totals
      class.probabilities[i,j] <- class.prob.totals$prob
      class.totals[i,j] <- class.prob.totals$total
    }
  }
  
  # Name class totals and class probabilities columns and rows
  rownames(class.probabilities) <- paste("C",c(1:ncol(offline.classes)),sep="")
  colnames(class.probabilities) <- paste("C",c(1:ncol(offline.classes)),sep="")
  rownames(class.totals) <- paste("C",c(1:ncol(offline.classes)),sep="")
  colnames(class.totals) <- paste("C",c(1:ncol(offline.classes)),sep="")
  
  result$class.totals <- class.totals
  result$class.probabilities <- class.probabilities
  
  return(result)
}
# -------------------------------------------------------------------------------------------------

# For each neuron, compute its average output
# -------------------------------------------------------------------------------------------------
getAverageNeuronOutputs <- function(output.som.map,num.micro.clusters){
  
  # Store the average outputs
  average.output.som.map <- vector(length=num.micro.clusters,mode="numeric")
  
  for(i in 1:num.micro.clusters){
    distances <- output.som.map$distances[as.numeric(which(output.som.map$unit.classif==i))]
    average.output.som.map[i] <- mean(sapply(distances,function(x){return(exp(-x))}))
  }
  
  return(average.output.som.map)
}
# -------------------------------------------------------------------------------------------------


# Computes radius factors for each neuron, based on the training examples these neurons classified
# -------------------------------------------------------------------------------------------------
computeRadiusFactorMC <- function(micro.clusters,output.som.map,data){
  
  for(i in 1:length(micro.clusters)){
    
    # Distances of codebook vectors to classified mapped instances
    distances <- output.som.map$distances[as.numeric(which(output.som.map$unit.classif==i))]
    
    # Indexes of examples mapped to neuron i
    indexes.data.mapped <- which(output.som.map$unit.classif==i)
    
    # Get max distance that classifies the farthest instance
    rFact <- max(distances)
    
    # Search for the most isolated point
    data.mapped <- data[indexes.data.mapped,,drop=F]
    
    # Centroid of the micro-cluster (weight vector)
    centroid <- micro.clusters[[i]][[1]]
    
    if(length(indexes.data.mapped) <= 1){browser()}
    
    # Get the nearest neighbour of each point
    d2 <- get.knn(data.mapped,k=1)
    
    # Search for isolated and its corresponding farthest neighbour
    isolated.index <- which(d2$nn.dist[,1]==max(d2$nn.dist[,1]))
    
    # If more than one, the isolated one will be the farthest from the centroid
    if(length(isolated.index) > 1){
      d3 <- as.matrix(dist(rbind(data.mapped[isolated.index,],centroid)))[length(isolated.index)+1,]
      isolated.index <- isolated.index[which(d3==max(d3))]
    }  
    
    # Position of the farthest.index index
    pos <- isolated.index
    
    # Search for nd.rFact 
    nd.rFact <- 9999999999999
    d2.sorted <- sort(d2$nn.dist[,1],decreasing=TRUE,index.return=TRUE)
    previous <- 0
    next.index <- 2
    checking <- 0
    
    while(nd.rFact >= rFact){
      checking <- checking + 1
    
      # Corresponding farthest neighbour
      farthest.index <- d2$nn.index[pos,]
      if(farthest.index == previous){ # Oh! We are in circles!
        next.index <- next.index + 1
        farthest.index <- d2.sorted$ix[next.index]
      }
      
      # Get distance between isolated.index and micro-cluster centroid
      nd.rFact <- euc.dist(data.mapped[farthest.index,],centroid)
      if(is.na(nd.rFact)){browser()}

      previous <- pos
      pos <- farthest.index
      
      if(checking==100){browser()}
    }
    
    if(nd.rFact == rFact){browser()}
    
    # Store radius factor and average output
    micro.clusters[[i]][[9]] <- rFact
    micro.clusters[[i]][[12]] <- nd.rFact 
    
    #plot(data.mapped)
    #points(centroid[1],centroid[2],col=2)
    #points(data.mapped[isolated.index,1],data.mapped[isolated.index,2],col=3)
    #points(data.mapped[farthest.index,1],data.mapped[farthest.index,2],col=4)
    #browser()
    
    if(rFact < nd.rFact){browser()}
    
    #plot(data.mapped, col=1, xlim=c(min(data.mapped[,1]),max(data.mapped[,1])), ylim=c(min(data.mapped[,2]),max(data.mapped[,2])), pch = 19, cex = 0.3)
    #points(micro.clusters[[i]][[1]][1],micro.clusters[[i]][[1]][2], col=2, pch = 19, cex = 0.5)
    #idmax <- which(max(distances) == distances)
    #segments(micro.clusters[[i]][[1]][1],micro.clusters[[i]][[1]][2],data.mapped[idmax,1],data.mapped[idmax,2])
    #segments(micro.clusters[[i]][[1]][1],micro.clusters[[i]][[1]][2],data.mapped[farthest.index,1],data.mapped[farthest.index,2])
    #browser()
  }
  
  return(micro.clusters)
}
# -------------------------------------------------------------------------------------------------


# For each neuron and each class, updates its conditional probabilitiy thresholds
# We will consider a p(x|y_j) = average.output.som.map for all neurons
# That means we are considering the output of the neurons to be average.output.som.map to calculate the threshold
# We consider that all classes are predicted to calculate the thresholds
# -------------------------------------------------------------------------------------------------
updateCondProbabilitiesNeurons <- function(som.micro.clusters,class.probabilities,z){
  
  z <- nrow(class.probabilities)
  #z <- ceiling(z)
  reduction.factor <- 1
  
  # go through all classes
  for(idx.class in 1:nrow(class.probabilities)){
    
    # Return the probabilities sorted
    class.prob <- class.probabilities[idx.class,]
    class.prob[idx.class] <- 0
    prob.sorted <- sort(class.prob,decreasing=TRUE,index.return=TRUE)
    idx.j <- idx.class
    prob.j <- class.probabilities[idx.j,idx.j] # p(y_j)
    
    prob.k.j <- 1
    for(k in 1:z){ # p(y_k|y_j)
      idx.k <- prob.sorted$ix[k]
      if(class.probabilities[idx.k,idx.j] > 0 & idx.k != idx.j){
        prob.k.j <- prob.k.j * class.probabilities[idx.k,idx.j]
      }  
    }
    
    # Go through all micro-clusters
    for(i in 1:length(som.micro.clusters[[idx.class]][[2]])){
      prob.x.j <- som.micro.clusters[[idx.class]][[2]][[i]][[11]] # p(x|y_j)
      prob.j.ks.x <- prob.j * prob.k.j * prob.x.j #p(y_j | y_k, x)
      
      som.micro.clusters[[idx.class]][[2]][[i]][[10]] <- prob.j.ks.x * reduction.factor
    }
  }
  
  return(som.micro.clusters)
}
# -------------------------------------------------------------------------------------------------

# For each neuron calculates its conditional probabilitiy thresholds for a given class
# We will consider a p(x|y_j) = average.output.som.map for all neurons
# That means we are considering the output of the neurons to be average.output.som.map to calculate the threshold
# We consider that all classes are predicted to calculate the thresholds
# -------------------------------------------------------------------------------------------------
getCondProbabilitiesNeurons <- function(micro.clusters,class.probabilities,average.output.som.map,idx.class,z){

  z <- nrow(class.probabilities)
  #z <- ceiling(z)
  reduction.factor <- 1
  
  # Return the probabilities sorted
  class.prob <- class.probabilities[idx.class,]
  class.prob[idx.class] <- 0
  prob.sorted <- sort(class.prob,decreasing=TRUE,index.return=TRUE)
  idx.j <- idx.class
  prob.j <- class.probabilities[idx.j,idx.j] # p(y_j)

  prob.k.j <- 1
  for(k in 1:z){ # p(y_k|y_j)
    idx.k <- prob.sorted$ix[k]
    if(class.probabilities[idx.k,idx.j] > 0 & idx.k != idx.j){
      prob.k.j <- prob.k.j * class.probabilities[idx.k,idx.j]
    }  
  }
  
  # Go through all micro-clusters
  for(i in 1:length(micro.clusters)){
    prob.x.j <- average.output.som.map[i] # p(x|y_j)
    prob.j.ks.x <- prob.j * prob.k.j * prob.x.j #p(y_j | y_k, x)
    
    micro.clusters[[i]][[10]] <- prob.j.ks.x * reduction.factor
    micro.clusters[[i]][[11]] <- average.output.som.map[i]
  }
  
  return(micro.clusters)
}
# -------------------------------------------------------------------------------------------------


# Compute label cardinality
# -------------------------------------------------------------------------------------
computeLabelCardinality <- function(offline.classes){
  
  z <- 0
  
  for(i in 1:nrow(offline.classes)){
    labels <- offline.classes[i,]
    idx.classified <- which(labels == 1)
    z <- z + length(idx.classified)
  }
  
  z <- z / nrow(offline.classes)
  
  return(z)
}

# -------------------------------------------------------------------------------------


# Compute micro-cluster properties
# -------------------------------------------------------------------------------------
computeMicroClusters <- function(som.map,class.id,min.ex){
  # Will store properties for each micro cluster
  # Each class will be represented by a set of neurons, whose mapped examples form micro-clusters
  # Properties: 1-LS, 2-SS, 3-n, 4-t, 5-y, 6-Radius, 7-time.stamp.creation, 8-class position in the matrices with properties,
  #             9-radius factor for novelty detection 1, 10-conditional probability threshold, 11-average output, 
  #             12-radius factor for novelty detection 2
  
  # Store results
  results <- list()
  
  # Get neurons to remove
  table.som.map.classif <- table(som.map$unit.classif)
  neurons <- as.numeric(names(table.som.map.classif[table.som.map.classif >= min.ex]))
  
  # Of course! We cannot eliminate all neurons!
  while(length(neurons) == 0){
    min.ex <- min.ex - 1
    neurons <- as.numeric(names(table.som.map.classif[table.som.map.classif >= min.ex]))
  }
  idx.neurons.to.remove <- as.numeric(names(table.som.map.classif[table.som.map.classif < min.ex]))
  
  # We cannot allow only two instances in a neuron! That generates an error!
  results$less.than.three <- 0
  if(min.ex < 3){
    results$less.than.three <- 1
  }
  
  # Mark neurons not used
  for(i in idx.neurons.to.remove){
    som.map$unit.classif[som.map$unit.classif==i] <- 0
  }
  
  # Structure for micro clusters
  micro.clusters <- lapply(1:length(neurons), function(x) as.list(rep(0,12)))
  
  # There may be empty neurons! Lets keep only the used ones
  som.map$codes[[1]] <- som.map$codes[[1]][neurons,,drop=F] 
  
  # Go through all neurons
  pos <- 0
  for(idx in neurons){
    indexes <- which(som.map$unit.classif==idx)
    pos <- pos + 1
    micro.clusters[[pos]][[1]] <- as.numeric(som.map$codes[[1]][pos,])
    
    micro.clusters[[pos]][[3]] <- length(indexes)
    micro.clusters[[pos]][[5]] <- class.id # class the micro-cluster represents
    micro.clusters[[pos]][[7]] <- 0 # Time stamp when created (Will not be modified)
    micro.clusters[[pos]][[8]] <- class.id # class position in the matrices with properties (Will not be modified)
    
    # We need to do that if we removed codebooks
    som.map$unit.classif[which(som.map$unit.classif==idx)] <- pos
  }
  
  # Remove the useless from som map
  #som.map$distances <- som.map$distances[-which(som.map$unit.classif==0)]
  #som.map$unit.classif <- som.map$unit.classif[-which(som.map$unit.classif==0)]
  
  results$som.map <- som.map
  results$micro.clusters <- micro.clusters
  
  return(results)
}
# -------------------------------------------------------------------------------------


# Get parameter values from file
# -------------------------------------------------------------------------------------
getParameterValues <- function(param.file){
  
  parameters <- c("train.mode","n0","n1","omega","num.iterations","theta","min.examples.cluster",
                  "train.data","test.data","num.classes","novel.classes","num.evaluation.windows","window.stm.check",
                  "num.experiments","empty.predictions","grid.dimension","update.model.info")
  
  values <- list()
  
  train.mode <- 0
  num.classes <- 0
  novel.classes <- 0
  n0 <- 0
  n1 <- 0
  grid.dimension <- 0
  num.iterations <- 0
  omega <- 0
  theta <- 0
  min.examples.cluster <- 0
  train.data <- 0
  test.data <- 0
  num.evaluation.windows <- 0
  window.stm.check <- 0
  num.experiments <- 0
  empty.predictions <- 0
  update.model.info <- 0
  
  # parameter file
  content <- scan(param.file,what="character",sep="\n")
  
  for(i in 1:length(parameters)){
    
    idx <- grep(parameters[i],content)
    value <- unlist(strsplit(content[idx],"="))[2]
    value <- gsub(" ","",value)
    
    if(parameters[i] == "n0"){
      n0 <- as.numeric(value)
    }
    else if(parameters[i] == "num.iterations"){
      num.iterations <- as.numeric(value)
    }
    else if(parameters[i] == "num.experiments"){
      num.experiments <- as.numeric(value)
    }
    else if(parameters[i] == "num.evaluation.windows"){
      num.evaluation.windows <- as.numeric(value)
    }
    else if(parameters[i] == "n1"){
      n1 <- as.numeric(value)
    }
    else if(parameters[i] == "train.data"){
      train.data <- value
    }
    else if(parameters[i] == "test.data"){
       test.data <- value
    }
    else if(parameters[i] == "train.mode"){
      train.mode <- value
    }
    else if(parameters[i] == "num.classes"){
      num.classes <- as.numeric(value)
    }
    else if(parameters[i] == "grid.dimension"){
      grid.dimension <- as.numeric(value)
    }
    else if(parameters[i] == "update.model.info"){
      update.model.info <- as.numeric(value)
    }
  }  
  
  values$n0 <- n0
  values$n1 <- n1
  values$num.iterations <- num.iterations
  values$train.data <- train.data
  values$test.data <- test.data
  values$num.classes <- num.classes
  values$train.mode <- train.mode
  values$num.evaluation.windows <- num.evaluation.windows
  values$num.experiments <- num.experiments
  values$grid.dimension <- grid.dimension
  values$update.model.info <- update.model.info
  
  return(values)
}
# -------------------------------------------------------------------------------------

# Compute probabilities p(i) and p(i|j)
# Also counts total of examples per classes and intersections
# -------------------------------------------------------------------------------------
# classes data frame, index of classes
get.probabilities <- function(classes,i,j){

	result <- list()
	result$prob <- 0
	result$total <- 0

	if(i == j){ # p(i)
		total.i <- length(which(classes[,i]==1))
		total <- nrow(classes)

		result$total <- total.i
		result$prob <- total.i / total
	}
	else{ # p(i|j)
		total.j <- length(which(classes[,j]==1))
		examples.i <- which(classes[,i]==1)
		examples.j <- which(classes[,j]==1)
		intersect.i.j <- intersect(examples.i,examples.j)
		total.i.j <- length(intersect.i.j)

		result$total <- total.i.j
		result$prob <- total.i.j / total.j
	}	

	return(result)
}
# -------------------------------------------------------------------------------------


# Euclidean distance
# -------------------------------------------------------------------------------------
euc.dist <- function(x, y){

    distance <- sqrt(sum((x - y)^2))

	return(distance)
}
# -------------------------------------------------------------------------------------

# Macro precision, recall, f-measure - calculated separated
# -------------------------------------------------------------------------------------
macro.precision.recall.fmeasure <- function(true,predicted){
  
  num.labels <- ncol(true)
  num.examples <- nrow(true)
  total.prec <- 0
  total.reca <- 0
  result <- list()
  
  for(j in 1:num.labels){
    tp <- 0
    fp <- 0
    fn <- 0
    prec <- 0
    reca <- 0
    for(i in 1:num.examples){
      if(true[i,j]==1 && predicted[i,j]==1){
        tp <- tp+1
      }
      else if(true[i,j]==0 && predicted[i,j]==1){
        fp <- fp+1
      }
      else if(true[i,j]==1 && predicted[i,j]==0){
        fn <- fn+1
      }
    }
    
    if((tp + fp) == 0){
      prec <- 0
    }
    else{
      prec <- tp / (tp + fp)
    }
    
    if((tp + fn) == 0){
      reca <- 0
    }
    else{
      reca <- tp / (tp + fn)
    }
    
    total.prec <- total.prec + prec
    total.reca <- total.reca + reca
  }
  
  result$ma.precision <- total.prec / num.labels
  result$ma.recall <- total.reca / num.labels
  result$ma.fmeasure <- (2 * result$ma.precision * result$ma.recall) / (result$ma.precision + result$ma.recall)
  
  return(result)	
}
# -------------------------------------------------------------------------------------

# Macro precision, recall, f-measure - calculated separated
# Calculatins separated for each evaluation window
# Considers novel classes
# -------------------------------------------------------------------------------------
macro.precision.recall.fmeasure.windows.NP <- function(true,predicted,num.evaluation.windows,novel.classes){
  
  num.labels.total <- ncol(true)
  num.labels <- ncol(true)-length(novel.classes)
  num.examples <- nrow(true)
  result <- list()
  ma.precision.window <- vector()
  ma.recall.window <- vector()
  ma.fmeasure.window <- vector()
  
  # Get positions of the first occurrence of the novel classes in the true data
  novel.classes.first.positions <- vector()
  true.NP <- true[,((ncol(true)-length(novel.classes))+1) : ncol(true),drop=F]
  for(i in 1:length(novel.classes)){
    pos <- match(1,true.NP[,i])
    novel.classes.first.positions <- c(novel.classes.first.positions,pos)
  }  
  
  # Number of examples per window. Try to keep an equal number of examples per window
  num.examples.window <- floor(num.examples / num.evaluation.windows)
  evaluation.windows <- rep(num.examples.window,num.evaluation.windows)
  rest <- nrow(predicted) - (num.examples.window*num.evaluation.windows)
  if(rest > 0){
    for(i in 1:rest){
      evaluation.windows[i] <- evaluation.windows[i]+1
    }
  }
  
  ini <- 1
  end <- 0
  tp <- vector(length=num.labels.total, mode="numeric")
  fp <- vector(length=num.labels.total, mode="numeric")
  fn <- vector(length=num.labels.total, mode="numeric")
  beta <- 1
  
  for(i in 1:num.evaluation.windows){
    end <- end + evaluation.windows[i]
    
    predictions.window <- predicted[ini:end,,drop=F]
    true.classes.window <- true[ini:end,,drop=F]
    num.examples.local <- nrow(predictions.window)
    
    total.prec <- 0
    total.reca <- 0
    total.fmeas <- 0
    
    for(j in 1:num.labels){
      
      prec <- 0
      reca <- 0
      fmeas <- 0
      
      # Index the number of each example evaluated
      index.example <- ini

      for(k in 1:num.examples.local){
        
        index.example <- index.example + 1
        if(is.na(match(index.example,novel.classes.first.positions)) == FALSE & j == 1){
          num.labels <- num.labels + 1
        }
        
        if(true.classes.window[k,j]==1 && predictions.window[k,j]==1){
          tp[j] <- tp[j]+1
        }
        else if(true.classes.window[k,j]==0 && predictions.window[k,j]==1){
          fp[j] <- fp[j]+1
        }
        else if(true.classes.window[k,j]==1 && predictions.window[k,j]==0){
          fn[j] <- fn[j]+1
        }
      }
      
      # Treat special cases just like Mulan
      # (i) if the algorithm outputs the empty set and the ground truth is the empty set,
      # then we consider precision equal to 1 (ii) if the algorithm outputs the
      # empty set and the ground truth is not empty, then we consider precision
      # equal to 0.
      if (tp[j] + fp[j] + fn[j] == 0) {
        prec <- 1
      }
      else if (tp[j] + fp[j] == 0) {
        prec <- 0
      }
      else{
        prec <- tp[j] / (tp[j] + fp[j])
      }
      
      # Treat special cases just like Mulan
      # (i) if the algorithm outputs the empty set and the ground truth is the empty set,
      # then we consider recall equal to 1 (ii) if the algorithm does not output
      # the empty set and the ground truth is empty, then we consider recall
      # equal to 0.  
      if (tp[j] + fp[j] + fn[j] == 0) {
        reca <- 1
      }
      else if (tp[j] + fn[j] == 0) {
        reca <- 0;
      }
      else{
        reca <- tp[j] / (tp[j] + fn[j]);
      } 
      
      # Treat special cases just like Mulan
      # (i) if the algorithm outputs the empty set and the ground truth is the empty
      # set, then we consider F equal to 1 (ii) if the algorithm outputs the
      # empty set and the ground truth is not empty, then we consider F equal to
      # 0 (iii) if the ground truth is empty and the algorithm does not output
      # the empty set, then we consider F equal to 0 (iv) if neither the ground
      # truth nor the algorithm's prediction is the empty set and their
      # intersection is empty, then we consider F equal to 0.
      if (tp[j] + fp[j] + fn[j] == 0) {
        fmeas <- 1
      }
      else{
        beta2 = beta * beta;
        fmeas <- ((beta2 + 1) * tp[j]) / ((beta2 + 1) * tp[j] + beta2 * fn[j] + fp[j])
      }
      
      total.prec <- total.prec + prec
      total.reca <- total.reca + reca
      total.fmeas <- total.fmeas + fmeas
    }
    
    precision <- total.prec / num.labels
    recall <- total.reca / num.labels
    fmeasure <- total.fmeas / num.labels
    
    ma.precision.window <- c(ma.precision.window,precision)
    ma.recall.window <- c(ma.recall.window,recall)
    ma.fmeasure.window <- c(ma.fmeasure.window,fmeasure)
    
    ini <- end + 1
  }

  result$ma.precision <- total.prec / num.labels
  result$ma.recall <- total.reca / num.labels
  result$ma.fmeasure <- total.fmeas / num.labels
  
  result$ma.precision.window <- ma.precision.window
  result$ma.recall.window <- ma.recall.window
  result$ma.fmeasure.window <- ma.fmeasure.window
  
  return(result)	
  
}
# -------------------------------------------------------------------------------------

# Macro precision, recall, f-measure - calculated separated
# Calculatins separated for each evaluation window
# Without considering novel classes
# -------------------------------------------------------------------------------------
macro.precision.recall.fmeasure.windows <- function(true,predicted,num.evaluation.windows){
  
  num.labels <- ncol(true)
  num.examples <- nrow(true)
  result <- list()
  ma.precision.window <- vector()
  ma.recall.window <- vector()
  ma.fmeasure.window <- vector()
  
  # Number of examples per window. Try to keep an equal number of examples per window
  num.examples.window <- floor(num.examples / num.evaluation.windows)
  evaluation.windows <- rep(num.examples.window,num.evaluation.windows)
  rest <- nrow(predicted) - (num.examples.window*num.evaluation.windows)
  if(rest > 0){
    for(i in 1:rest){
      evaluation.windows[i] <- evaluation.windows[i]+1
    }
  }
  
  ini <- 1
  end <- 0
  tp <- vector(length=num.labels, mode="numeric")
  fp <- vector(length=num.labels, mode="numeric")
  fn <- vector(length=num.labels, mode="numeric")
  beta <- 1
  
  for(i in 1:num.evaluation.windows){
    end <- end + evaluation.windows[i]
    
    predictions.window <- predicted[ini:end,,drop=F]
    true.classes.window <- true[ini:end,,drop=F]
    num.examples.local <- nrow(predictions.window)
    
    total.prec <- 0
    total.reca <- 0
    total.fmeas <- 0
    
    for(j in 1:num.labels){
      
      prec <- 0
      reca <- 0
      fmeas <- 0
      
      # Index the number of each example evaluated
      index.example <- ini
      
      for(k in 1:num.examples.local){
        
        index.example <- index.example + 1
        
        if(true.classes.window[k,j]==1 && predictions.window[k,j]==1){
          tp[j] <- tp[j]+1
        }
        else if(true.classes.window[k,j]==0 && predictions.window[k,j]==1){
          fp[j] <- fp[j]+1
        }
        else if(true.classes.window[k,j]==1 && predictions.window[k,j]==0){
          fn[j] <- fn[j]+1
        }
      }
      
      # Treat special cases just like Mulan
      # (i) if the algorithm outputs the empty set and the ground truth is the empty set,
      # then we consider precision equal to 1 (ii) if the algorithm outputs the
      # empty set and the ground truth is not empty, then we consider precision
      # equal to 0.
      if (tp[j] + fp[j] + fn[j] == 0) {
        prec <- 1
      }
      else if (tp[j] + fp[j] == 0) {
        prec <- 0
      }
      else{
        prec <- tp[j] / (tp[j] + fp[j])
      }
      
      # Treat special cases just like Mulan
      # (i) if the algorithm outputs the empty set and the ground truth is the empty set,
      # then we consider recall equal to 1 (ii) if the algorithm does not output
      # the empty set and the ground truth is empty, then we consider recall
      # equal to 0.  
      if (tp[j] + fp[j] + fn[j] == 0) {
        reca <- 1
      }
      else if (tp[j] + fn[j] == 0) {
        reca <- 0
      }
      else{
        reca <- tp[j] / (tp[j] + fn[j])
      } 
      
      # Treat special cases just like Mulan
      # (i) if the algorithm outputs the empty set and the ground truth is the empty
      # set, then we consider F equal to 1 (ii) if the algorithm outputs the
      # empty set and the ground truth is not empty, then we consider F equal to
      # 0 (iii) if the ground truth is empty and the algorithm does not output
      # the empty set, then we consider F equal to 0 (iv) if neither the ground
      # truth nor the algorithm's prediction is the empty set and their
      # intersection is empty, then we consider F equal to 0.
      if (tp[j] + fp[j] + fn[j] == 0) {
        fmeas <- 1
      }
      else{
        beta2 = beta * beta
        fmeas <- ((beta2 + 1) * tp[j]) / ((beta2 + 1) * tp[j] + beta2 * fn[j] + fp[j])
      }
      
      total.prec <- total.prec + prec
      total.reca <- total.reca + reca
      total.fmeas <- total.fmeas + fmeas
    }
    
    precision <- total.prec / num.labels
    recall <- total.reca / num.labels
    fmeasure <- total.fmeas / num.labels
    
    ma.precision.window <- c(ma.precision.window,precision)
    ma.recall.window <- c(ma.recall.window,recall)
    ma.fmeasure.window <- c(ma.fmeasure.window,fmeasure)
    
    ini <- end + 1
  }
  
  result$ma.precision <- total.prec / num.labels
  result$ma.recall <- total.reca / num.labels
  result$ma.fmeasure <- total.fmeas / num.labels
  
  result$ma.precision.window <- ma.precision.window
  result$ma.recall.window <- ma.recall.window
  result$ma.fmeasure.window <- ma.fmeasure.window
  
  return(result)	
  
}
# -------------------------------------------------------------------------------------


# Calculates the unknown rate per window
# -------------------------------------------------------------------------------------
unknown.rate.windows <- function(all.unclassified.indexes,time.stamp.stm.removed,true,num.evaluation.windows,novel.classes){
  
  num.labels.total <- ncol(true)
  num.labels.ini <- ncol(true)-length(novel.classes)
  num.examples <- nrow(true)
  result <- list()
  
  # Get positions of the first occurrence of the novel classes in the true data
  novel.classes.first.positions <- vector()
  true.NP <- true[,((ncol(true)-length(novel.classes))+1) : ncol(true),drop=F]
  for(i in 1:length(novel.classes)){
    pos <- match(1,true.NP[,i])
    novel.classes.first.positions <- c(novel.classes.first.positions,pos)
  }  
  
  # Number of examples per window. Try to keep an equal number of examples per window
  num.examples.window <- floor(num.examples / num.evaluation.windows)
  evaluation.windows <- rep(num.examples.window,num.evaluation.windows)
  rest <- nrow(true) - (num.examples.window*num.evaluation.windows)
  if(rest > 0){
    for(i in 1:rest){
      evaluation.windows[i] <- evaluation.windows[i]+1
    }
  }
  
  ini <- 1
  end <- 0
  unknown.rate.window <- vector()
  num.unknown.window <- vector()
  UNKj <- vector(length=num.labels.total, mode="numeric")

  for(i in 1:num.evaluation.windows){
    end <- end + evaluation.windows[i]
    UnkRMj <- 0
    UnkRM <- 0
    # Are there unknown indexes in this window?
    #idx.unknown <- intersect(all.unclassified.indexes,seq(from=ini,to=end))
    idx.unknown <- all.unclassified.indexes[c(which(time.stamp.stm.removed > end),which(time.stamp.stm.removed == 0))]
    idx.unknown <- idx.unknown[idx.unknown<=end]
    
    if(length(idx.unknown) > 0){
      
      #true.classes.window <- true[min(idx.unknown):max(idx.unknown),,drop=F]
      true.classes.window <- true[ini:end,,drop=F]
      num.examples.local <- nrow(true.classes.window)
      num.labels <- num.labels.ini
      
      for(j in 1:num.labels){
        
        # Index the number of each example evaluated
        index.example <- ini
        total.examples.j <- 0
        
        # Number of unknown examples from class j
        row.names <- as.numeric(rownames(true.classes.window))
        idx.unknown.window <- which(row.names %in% idx.unknown)
        UNKj[j] <- UNKj[j] + length(which(true.classes.window[idx.unknown.window,j]==1))
        for(k in 1:num.examples.local){
          
          index.example <- index.example + 1
          if(is.na(match(index.example,novel.classes.first.positions)) == FALSE & j == 1){
            num.labels <- num.labels + 1
          }
          
          if(true.classes.window[k,j]==1){
            total.examples.j <- total.examples.j + 1
          }
        }
        
        # Total number of examples from class j
        Nj <- total.examples.j
        
        if(Nj > 0){
          UnkRMj <- UnkRMj + (UNKj[j] / Nj)
        }
      }
      UnkRM <- UnkRMj / num.labels
    }
    
    unknown.rate.window <- c(unknown.rate.window,UnkRM)
    num.unknown.window <- c(num.unknown.window,length(idx.unknown))

    #ini <- end + 1
  }
  
  result$unknown.rate.window <- unknown.rate.window
  result$num.unknown.window <- num.unknown.window

  return(result)	
}
# -------------------------------------------------------------------------------------


# Update class totals and class probabilities matrices
# -------------------------------------------------------------------------------------
updateClassTotalsProbabilities <- function(mapping,pred,num.pred,predictions,initial.number.classes,num.offline.instances){
  
  # Total number of instances explained
  mapping$total.instances <- mapping$total.instances + num.pred
  
  # Total number of examples predicted
  num.predictions <- mapping$total.instances - num.offline.instances
  
  if(nrow(pred) > 0){
    for(r in 1:nrow(pred)){
      # Neurons which explained the example
      idx.classes <- which(pred[r,] > 0)
      
      if(length(idx.classes) > 0){
        # Class totals
        for(i in 1:length(idx.classes)){
          for(j in 1:length(idx.classes)){
            idx.i <- idx.classes[i]
            idx.j <- idx.classes[j]
            if( pred[r,idx.i] == 1 | pred[r,idx.j] == 1 ){ 
              mapping$class.totals[idx.i,idx.j] <- mapping$class.totals[idx.i,idx.j] + 1
              # Otherwise we have already counted Totals[c_i,c_i]
            }
          }
        }
      }
    }
  }
  
  # We need to check if we are dealing with an existing class or witn a NP
  # The probabilities are calculated differently
  for(idx.i in 1:nrow(mapping$class.totals)){
    for(idx.j in 1:nrow(mapping$class.totals)){

        if(idx.i == idx.j){
          mapping$class.probabilities[idx.i,idx.j] <- mapping$class.totals[idx.i,idx.j] / mapping$total.instances
        }
        else{
          mapping$class.probabilities[idx.i,idx.j] <- mapping$class.totals[idx.i,idx.j] / mapping$class.totals[idx.j,idx.j]
        }
      }
  }
  
  return(mapping)
}
# -------------------------------------------------------------------------------------


# Update model information when an example is explained
# -------------------------------------------------------------------------------------
update.model.information <- function(mapping,pred,x,time.stamp,n0,initial.number.classes,predictions,winner,inst.l){
	
  # Micro clusters and codebooks
  
  # Neurons which explained the example
  idx.classes <- which(pred == 1)
  
  for(idx in idx.classes){
    
    # Number of neurons
    #num.neurons <- ncol(winner[[idx]]$nn.index)
    
    #for(idx.n in 1:num.neurons){
      
      idx.neuron <- winner[[idx]]$nn.index[inst.l,1]
      
      # Micro clusters
      #mapping$som.micro.clusters[[idx]][[2]][[idx.neuron]][[1]] <-  mapping$som.micro.clusters[[idx]][[2]][[idx.neuron]][[1]] + x
      #mapping$som.micro.clusters[[idx]][[2]][[idx.neuron]][[2]] <-  mapping$som.micro.clusters[[idx]][[2]][[idx.neuron]][[2]] + x*x
      mapping$som.micro.clusters[[idx]][[2]][[idx.neuron]][[3]] <-  mapping$som.micro.clusters[[idx]][[2]][[idx.neuron]][[3]] + 1
      mapping$som.micro.clusters[[idx]][[2]][[idx.neuron]][[4]] <- time.stamp
      
      # Codebooks
      delta <- n0 * (x - mapping$som.micro.clusters[[idx]][[1]]$codes[[1]][idx.neuron,]) * exp(-winner[[idx]]$nn.dist[inst.l,1])
      old <- mapping$som.micro.clusters[[idx]][[1]]$codes[[1]][idx.neuron,]
      mapping$som.micro.clusters[[idx]][[1]]$codes[[1]][idx.neuron,] <- old + delta
      mapping$som.micro.clusters[[idx]][[2]][[idx.neuron]][[1]] <- old + delta
      
      #computeRadiusFactorMC(micro.clusters,output.som.map,average.output.som.map,data){
      
      #old <- mapping$som.micro.clusters[[idx]][[1]]$codes[[1]][idx.neuron,]
      #mapping$som.micro.clusters[[idx]][[1]]$codes[[1]][idx.neuron,] <- old*(1-n0) + x*n0
      
      #LS <- mapping$som.micro.clusters[[idx]][[2]][[idx.neuron]][[1]]
      #SS <- mapping$som.micro.clusters[[idx]][[2]][[idx.neuron]][[2]]
      #N <- mapping$som.micro.clusters[[idx]][[2]][[idx.neuron]][[3]]
      #mapping$som.micro.clusters[[idx]][[2]][[idx.neuron]][[6]] <- getStandardDeviation(LS,SS,N)
    #}
  }
  
  # Calculates the largest micro-cluster radius for all classes
  # mapping$largest.radius <- getLargestRadiusMC(mapping$som.micro.clusters,idx.classes,mapping$largest.radius)
  
	return(mapping)
}
# -------------------------------------------------------------------------------------


# Classification using winning neuron and conditional probabilities 
# Uses probabilities considering the Bayes Rule and label cardinality z
# -------------------------------------------------------------------------------------
predict.kohonen.win.bayes <- function(som.micro.clusters,dataset,class.probabilities,z){

  # number of examples
  num.examples <- nrow(dataset)
  
  # Store predictions
  predictions <- matrix(nrow=nrow(dataset), ncol=nrow(class.probabilities), data=0)

  step = 100
  for(inst in seq(from=1, to=num.examples, by=step)){
    
    matrix.examples <- 0
    if(inst+step > num.examples){
      matrix.examples <- dataset[inst:num.examples,,drop=F]
      rownames(matrix.examples) <- inst:num.examples
    }
    else{
      matrix.examples <- dataset[inst:(inst+(step-1)),,drop=F]
      rownames(matrix.examples) <- inst:(inst+(step-1))
    } 
    
    rownames.matrix.examples <- as.numeric(rownames(matrix.examples))
    num.examples.matrix.examples <- nrow(matrix.examples)
    
    # Num patterns (sets of neurons) to be considered in the chunck of examples (matrix.examples)
    num.patterns <- nrow(class.probabilities)
    
    # Number of predictions for the chunck of examples
    num.predictions.matrix.examples <- 0

    # Get k nearest neighbour neurons for the set of examples
    winner <- list()
    all.n.ks <- vector()
    for(j in 1:num.patterns){
      #n.k <- ceiling(nrow(mapping$som.micro.clusters[[j]][[1]]$codes[[1]])*0.1)
      n.k <- nrow(som.micro.clusters[[j]][[1]]$codes[[1]])
      if(n.k %% 2 == 0){n.k <- n.k - 1} # Guarantee odd number
      if(nrow(som.micro.clusters[[j]][[1]]$codes[[1]])==1){n.k <- 1}
      winner[[j]] <- get.knnx(som.micro.clusters[[j]][[1]]$codes[[1]],matrix.examples,k=n.k)
      all.n.ks <- c(all.n.ks,n.k)
    } 

    # For each example
    for(inst.l in 1:num.examples.matrix.examples){
      
      # Get incoming example
      x <- as.numeric(matrix.examples[inst.l,])
      
      # Get winning neuron
      outputs <- vector() 
      outputs.prob1 <- vector()
      distances <- vector()
      winners.index <- vector()
      
      # Structures for majority voting
      vet.distances <- vector()
      vet.indexes <- vector()

      # For each class, get the winning neuron among the k closest neurons
      for(j in 1:num.patterns){
        
        # winning neuron
        win <- winner[[j]]$nn.index[inst.l,1]
        
        # Centroid of the winning neuron
        centroid <- som.micro.clusters[[j]][[2]][[win]][[1]] #/ som.micro.clusters[[j]][[2]][[win]][[3]]
        
        # Euclidean distance between centroids of the nearest neuron (micro-cluster) and example
        distance <- winner[[j]]$nn.dist[inst.l,1]
        #distance <- euc.dist(centroid,x)
        
        # p(x|y_j) considering the winning neuron
        prob1 <- exp(-winner[[j]]$nn.dist[inst.l,1]) 
        
        # Class (pos) the neuron represents
        pos <- som.micro.clusters[[j]][[2]][[win]][[8]]
        
        # p(y_j|x) = p(y_j)*p(x|y_j) considering the winning neuron
        prob2 <- class.probabilities[pos,pos]*prob1  
        
        winners.index <- c(winners.index,win)
        distances <- c(distances,distance)
        outputs.prob1 <- c(outputs.prob1,prob1) 
        outputs <- c(outputs,prob2)
        
        # Storing all distances and indexes of classes
        # This will be used to decide winner class
        vet.distances <- c(vet.distances,winner[[j]]$nn.dist[inst.l,])
        vet.indexes <- c(vet.indexes,rep(j,ncol(winner[[j]]$nn.dist)))
      }
  
      # Sort output values and return sorted indexes
      prob.sorted <- sort.winning.neurons(distances,vet.distances,vet.indexes,min(all.n.ks),num.patterns,winner,inst.l)
      idx.prob.sorted <- prob.sorted$ix
      
      #cat(rownames.matrix.examples[inst.l]," - ",outputs.prob1[idx.prob.sorted[1]]," - ",distances[idx.prob.sorted[1]]," - ",distances.codebook[idx.prob.sorted[1]]," - " ,mapping$som.micro.clusters[[idx.prob.sorted[1]]][[2]][[winners.index[idx.prob.sorted[1]]]][[6]] * radiusFactor,"\n")
      
      # Ceiling label cardinality 
      z <- ceiling(z)
      for(j in 1:z){
      #for(j in 1:ncol(class.probabilities)){
        if(j == 1){
          #cat("\nKohonen.R 1 ----",winners.index[idx.prob.sorted[j]],"--->",length(mapping$som.micro.clusters[[idx.prob.sorted[j]]][[2]][[winners.index[idx.prob.sorted[j]]]]),"\n")
          pos <- som.micro.clusters[[idx.prob.sorted[j]]][[2]][[winners.index[idx.prob.sorted[j]]]][[8]]
          predictions[rownames.matrix.examples[inst.l],pos] <- 1
          explained <- 1
        }
        else{
          #cat("\nKohonen.R 2 ----",winners.index[idx.prob.sorted[j]],"--->",length(mapping$som.micro.clusters[[idx.prob.sorted[j]]][[2]][[winners.index[idx.prob.sorted[j]]]]),"\n")
          idx.j <- som.micro.clusters[[idx.prob.sorted[j]]][[2]][[winners.index[idx.prob.sorted[j]]]][[8]]
          prob.j <- class.probabilities[idx.j,idx.j] # p(y_j)
          prob.x.j <- outputs.prob1[idx.prob.sorted[j]] # p(x|y_j)
          prob.k.j <- 1
          for(k in 1:(j-1)){ # p(y_k|y_j)
            #cat("\nKohonen.R 3 ----",winners.index[idx.prob.sorted[k]],"--->",length(mapping$som.micro.clusters[[idx.prob.sorted[k]]][[2]][[winners.index[idx.prob.sorted[k]]]]),"\n")
            idx.k <- som.micro.clusters[[idx.prob.sorted[k]]][[2]][[winners.index[idx.prob.sorted[k]]]][[8]]
            if(predictions[rownames.matrix.examples[inst.l],pos] == 1){
              prob.k.j <- prob.k.j * class.probabilities[idx.k,idx.j]
            }
          }
          prob.j.ks.x <- prob.j * prob.k.j * prob.x.j #p(y_j | y_k, x)
          
          #cat("\n Fora ND: ",prob.j.ks.x,"\n")
          cond.prob.thres <- som.micro.clusters[[idx.prob.sorted[j]]][[2]][[winners.index[idx.prob.sorted[j]]]][[10]] 
          if(prob.j.ks.x >= cond.prob.thres){ # Threshold!
            #cat("\nKohonen.R 4 ----",winners.index[idx.prob.sorted[j]],"--->",length(mapping$som.micro.clusters[[idx.prob.sorted[j]]][[2]][[winners.index[idx.prob.sorted[j]]]]),"\n")
            pos <- som.micro.clusters[[idx.prob.sorted[j]]][[2]][[winners.index[idx.prob.sorted[j]]]][[8]]
            predictions[rownames.matrix.examples[inst.l],pos] <- 1
          }
        }
      }
    }
  }  
  
 	return(predictions)
}
# -------------------------------------------------------------------------------------

# Get Variance Vector of a Micro Cluster - MOA implementation
# -------------------------------------------------------------------------------------
getVarianceVector <- function(LS,SS,N){
	
	EPSILON = 0.00005
	MIN_VARIANCE = 1e-50

	res <- vector()
	
	for(i in 1:length(LS)){
		ls <- LS[i]
		ss <- SS[i]

		lsDivN <- ls / N
		lsDivNSquared <- lsDivN * lsDivN
		ssDivN <- ss / N
		result <- ssDivN - lsDivNSquared
		res <- c(res,result)

		# Due to numerical errors, small negative values can occur.
    # We correct this by settings them to almost zero.
		if(res[i] <= 0){
			if(res[i] > -EPSILON){
				res[i] <- MIN_VARIANCE
			}
		}
	}	
	
	return(res)
}
# -------------------------------------------------------------------------------------



# Get Standard Deviation of a Micro Cluster
# Aggarwal, C. C., Yu, P. S., Han, J., & Wang, J. (2003). 
# A Framework for Clustering Evolving Data Streams. 
# Proceedings 2003 VLDB Conference, 81–92. doi:10.1016/b978-012722442-8/50016-1 
# -------------------------------------------------------------------------------------
getStandardDeviation <- function(LS,SS,N){
	
	variance.vector <- getVarianceVector(LS,SS,N)
	sumOfDeviation <- 0
	for(i in 1:length(variance.vector)){
		sumOfDeviation <- sumOfDeviation + variance.vector[i]
	}

	sumOfDeviation.sqrt <- sqrt(sumOfDeviation)

	return(sumOfDeviation.sqrt)
}
# -------------------------------------------------------------------------------------


# Calculate silhouette of a micro-cluster
# -------------------------------------------------------------------------------------
getSilhoutte <- function(b,a){
  
  silhou <- (b-a)/(max(b,a))
  
  return(silhou)
}
# -------------------------------------------------------------------------------------

# Given a codebook i, return its nearest codebook from a set of codebooks
# -------------------------------------------------------------------------------------
getNearest <- function(index,codes,group.NP.micro.clusters,new.micro.clusters){

  found <- 0
  minimum <- 9999999999999
  centroid.mc.index <- new.micro.clusters[[index]][[1]] / new.micro.clusters[[index]][[3]]
  for(i in 1:length(group.NP.micro.clusters)){
    for(j in 1:length(group.NP.micro.clusters[[i]])){
      idx <- group.NP.micro.clusters[[i]][j]
      #radius <- max(new.micro.clusters[[index]][[6]],new.micro.clusters[[idx]][[6]])*radius.factor
      radius <- new.micro.clusters[[idx]][[6]] * new.micro.clusters[[idx]][[9]]
      centroid.mc.idx <- new.micro.clusters[[idx]][[1]] / new.micro.clusters[[idx]][[3]]
      #euclidean.dist <- euc.dist(codes[index,],codes[idx,])
      euclidean.dist <- euc.dist(centroid.mc.index,centroid.mc.idx)
      if(euclidean.dist <= radius){
        if(euclidean.dist < minimum){
          minimum <- euclidean.dist
          group.NP.micro.clusters[[i]] <- c(group.NP.micro.clusters[[i]],index)
          found <- 1
        }  
      }
    }
  }
  
  if(found == 0){
    group.NP.micro.clusters[[length(group.NP.micro.clusters)+1]] <- index
  }
    
  return(group.NP.micro.clusters)
}
# -------------------------------------------------------------------------------------

# Verify distances between codebook vectors to see if they can represent a same NP
# -------------------------------------------------------------------------------------
joinNP <- function(som.map,new.micro.clusters){
  
  group.NP.micro.clusters <- list()
  vet <- 1
  group.NP.micro.clusters[[1]] <- vet
  
  if(length(new.micro.clusters) > 1){
    for(i in 2:length(new.micro.clusters)){
      group.NP.micro.clusters <- getNearest(i,som.map$codes[[1]],group.NP.micro.clusters,new.micro.clusters)
    }
  }
  
  return(group.NP.micro.clusters)
}
# -------------------------------------------------------------------------------------


# Include a micro-cluster and codebooks as extension of existing model
# -------------------------------------------------------------------------------------
includeMicroClusterExtension <- function(mapping,new.micro.clusters,closest.sorted,time.stamp,neuron.pos,som.map,i,j){
  
  result <- list()
  
  # Include all neurons
  for(n.pos in neuron.pos){
    
    # include micro-cluster
    num.existing.neurons <- length(mapping$som.micro.clusters[[closest.sorted$ix[j]]][[2]])
    
    mapping$som.micro.clusters[[closest.sorted$ix[j]]][[2]][[num.existing.neurons+1]] <- new.micro.clusters[[n.pos]]
    # Update micro-cluster properties (class [5], time-stamp creation [7], position in matrix [8] )
    
    # class
    class <- mapping$som.micro.clusters[[closest.sorted$ix[j]]][[2]][[num.existing.neurons]][[5]]
    mapping$som.micro.clusters[[closest.sorted$ix[j]]][[2]][[num.existing.neurons+1]][[5]] <- class
    
    # time stamp
    mapping$som.micro.clusters[[closest.sorted$ix[j]]][[2]][[num.existing.neurons+1]][[4]] <- time.stamp
    
    # time stamp creation
    mapping$som.micro.clusters[[closest.sorted$ix[j]]][[2]][[num.existing.neurons+1]][[7]] <- time.stamp
    
    # position in matrix with properties
    pos.matrix <- mapping$som.micro.clusters[[closest.sorted$ix[j]]][[2]][[num.existing.neurons]][[8]]
    mapping$som.micro.clusters[[closest.sorted$ix[j]]][[2]][[num.existing.neurons+1]][[8]] <- pos.matrix
    
    # include codebook vector
    old <- mapping$som.micro.clusters[[closest.sorted$ix[j]]][[1]]$codes[[1]]
    mapping$som.micro.clusters[[closest.sorted$ix[j]]][[1]]$codes[[1]] <- rbind(old,som.map$codes[[1]][n.pos,,drop=F])
  }
  
  result$mapping <- mapping
  result$pos.matrix <- pos.matrix
  
  return(result)
}
# -------------------------------------------------------------------------------------

# get indexes of instances classified by given neurons (mc.ids)
# -------------------------------------------------------------------------------------
get.idx.ids <- function(som.map,mc.ids){

  idx.i <- vector()
  
  for(i in mc.ids){
    idx.i <- c(idx.i,as.numeric(which(som.map$unit.classif==i)))
  }
  
  return(idx.i)
}
# -------------------------------------------------------------------------------------



# Auxiliar function to remove old neurons
# -------------------------------------------------------------------------------------
find.neurons.keep <- function(list.neurons,time.stamp,omega){
  
  keep <- FALSE
  if((time.stamp - list.neurons[[4]]) <= omega){
    keep <- TRUE
  }
  return(keep)
}
# -------------------------------------------------------------------------------------


# Count occurrences of NPs together with novel classes. For evaluation!
# -------------------------------------------------------------------------------------
countOccurrences <- function(predictions,true.classes,p,t){

  result <- list()
  
  first.occurrence <- 0 # The first time a NPi appears together with a class y_new
  a <- 0 # the number of times a NPi appears together with a class y_new
  b <- 0 # the number of times a class y_new appears, but a NPi does not appear
  c <- 0 # the number of times a NPi appears, but a class y_new does not appear
  d <- 0 # the number of times a NPi does not appear and a class y_new does not appear either
  
  for(i in 1:nrow(predictions)){
    if(predictions[i,p] == 1 & true.classes[i,t] == 1){
      a <- a + 1
      if(first.occurrence  == 0){
        first.occurrence <- i
      }  
    }
    else if(predictions[i,p] == 0 & true.classes[i,t] == 1){
      b <- b + 1
    }
    else if(predictions[i,p] == 1 & true.classes[i,t] == 0){
      c <- c + 1
    }
    else if(predictions[i,p] == 0 & true.classes[i,t] == 0){
      d <- d + 1
    }
  }
  
  result$a <- a
  result$b <- b
  result$c <- c
  result$d <- d
  result$first.occurrence <- first.occurrence
  
  return(result)
  
}
# -------------------------------------------------------------------------------------

# Computes F1 values for association between NPs and classes
# -------------------------------------------------------------------------------------
get.f1 <- function(counts){
  
  f1 <- 0
  beta <- 1
  
  if(counts$a == 0){
    f1 <- 0
  }
  else if( (counts$a + counts$b + counts$c) == 0){
    f1 <- 0
  }
  else{
    beta2 = beta * beta
    f1 <- ((beta2 + 1) * counts$a) / ((beta2 + 1) * counts$a + beta2 * counts$b + counts$c)
    #f1 <- counts$a / (counts$a + ((counts$b+counts$c)/2) )
  }
  
  return(f1)
}
# -------------------------------------------------------------------------------------













